@@ -23,28 +23,38 @@ from mcp.server.streamable_http import (
 )
 from mcp.shared.message import ServerMessageMetadata
 from rbt.mcp.v1.session_rbt import Session
-from reboot.aio.contexts import WorkflowContext
+from rbt.v1alpha1.errors_pb2 import StateNotConstructed
+from reboot.aio.applications import Application
+from reboot.aio.contexts import EffectValidation, WorkflowContext
 from reboot.aio.external import ExternalContext
 from reboot.aio.types import StateRef
-from reboot.mcp.event_store import DurableEventStore, replay
+from reboot.aio.workflows import at_least_once
+from reboot.mcp.event_store import (
+    DurableEventStore,
+    replay,
+    qualified_stream_id,
+)
 from reboot.mcp.servicers.session import (
     SessionServicer,
     _servers,
     _context,
 )
-from reboot.aio.workflows import at_least_once
 from reboot.mcp.servicers.stream import StreamServicer
 from reboot.std.collections.v1 import sorted_map
 from rebootdev.aio.headers import CONSENSUS_ID_HEADER, STATE_REF_HEADER
+from rebootdev.aio.backoff import Backoff
 from rebootdev.memoize.v1.memoize_rbt import Memoize
+from rebootdev.settings import DOCS_BASE_URL
 from starlette.applications import Starlette
+from starlette.datastructures import MutableHeaders
 from starlette.requests import Request
 from starlette.responses import Response, StreamingResponse
 from starlette.routing import Route
 from starlette.types import Receive, Scope, Send
 from types import MethodType
-from typing import Any, Callable, Literal, Protocol, cast
+from typing import Any, Callable, Literal, Protocol, TypeAlias, cast
 from uuid import uuid4, uuid5
+from uuid7 import create as uuid7  # type: ignore[import-untyped]
 
 logger = get_logger(__name__)
 
@@ -102,8 +112,10 @@ class DurableSession:
                     # TODO: figure out why `mypy` requires
                     # passing `method` which has a default.
                     method="notifications/resources/list_changed",
-                    _meta=mcp.types.NotificationParams.Meta(
-                        rebootEventId=str(event_id),
+                    params=mcp.types.NotificationParams(
+                        _meta=mcp.types.NotificationParams.Meta(
+                            rebootEventId=str(event_id),
+                        ),
                     ),
                 ),
             ),
@@ -126,8 +138,10 @@ class DurableSession:
                     # TODO: figure out why `mypy` requires
                     # passing `method` which has a default.
                     method="notifications/tools/list_changed",
-                    _meta=mcp.types.NotificationParams.Meta(
-                        rebootEventId=str(event_id),
+                    params=mcp.types.NotificationParams(
+                        _meta=mcp.types.NotificationParams.Meta(
+                            rebootEventId=str(event_id),
+                        ),
                     ),
                 ),
             ),
@@ -150,15 +164,17 @@ class DurableSession:
                     # TODO: figure out why `mypy` requires
                     # passing `method` which has a default.
                     method="notifications/prompts/list_changed",
-                    _meta=mcp.types.NotificationParams.Meta(
-                        rebootEventId=str(event_id),
+                    params=mcp.types.NotificationParams(
+                        _meta=mcp.types.NotificationParams.Meta(
+                            rebootEventId=str(event_id),
+                        ),
                     ),
                 ),
             ),
         )
 
 
-class ToolContextProtocol(Protocol):
+class DurableContextProtocol(Protocol):
 
     _event_aliases: set[str]
 
@@ -248,7 +264,7 @@ class ToolContextProtocol(Protocol):
          ...
 
 
-class ToolContext(WorkflowContext, ToolContextProtocol):
+class DurableContext(WorkflowContext, DurableContextProtocol):
 
     pass
 
@@ -281,8 +297,20 @@ class Tool:
     structured_output: bool | None
 
 
-class DurableMCP(fastmcp.FastMCP):
+LogLevel: TypeAlias = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
 
+
+class DurableMCP:
+    """
+    Proxy for `fastmcp.FastMCP`, but wrapping tools, prompts, and
+    resources appropriately to make them durable.
+
+    NOTE: this class explicitly does NOT extend from `fastmcp.FastMCP`
+    so that we don't mislead users into thinking some features are
+    implemented that are not.
+    """
+
+    _log_level: LogLevel
     _resources: list[Resource]
     _prompts: list[Prompt]
     _tools: list[Tool]
@@ -291,10 +319,9 @@ class DurableMCP(fastmcp.FastMCP):
         self,
         *,
         path: str,
-        log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "WARNING",
+        log_level: LogLevel = "WARNING",
     ):
-        super().__init__(log_level=log_level)
-
+        self._log_level = log_level
         self._path = path
         self._resources = []
         self._prompts = []
@@ -318,7 +345,43 @@ class DurableMCP(fastmcp.FastMCP):
         description: str | None = None,
         mime_type: str | None = None,
     ) -> Callable[[mcp.types.AnyFunction], mcp.types.AnyFunction]:
-        """Overrides `FastMCP.add_resource`."""
+        """Decorator to register a function as a resource.
+
+        The function will be called when the resource is read to generate its content.
+        The function can return:
+        - str for text content
+        - bytes for binary content
+        - other types will be converted to JSON
+
+        If the URI contains parameters (e.g. "resource://{param}") or the function
+        has parameters, it will be registered as a template resource.
+
+        Args:
+            uri: URI for the resource (e.g. "resource://my-resource" or "resource://{param}")
+            name: Optional name for the resource
+            title: Optional human-readable title for the resource
+            description: Optional description of the resource
+            mime_type: Optional MIME type for the resource
+
+        Example:
+            @server.resource("resource://my-resource")
+            def get_data() -> str:
+                return "Hello, world!"
+
+            @server.resource("resource://my-resource")
+            async get_data() -> str:
+                data = await fetch_data()
+                return f"Hello, world! {data}"
+
+            @server.resource("resource://{city}/weather")
+            def get_weather(city: str) -> str:
+                return f"Weather for {city}"
+
+            @server.resource("resource://{city}/weather")
+            async def get_weather(city: str) -> str:
+                data = await fetch_weather(city)
+                return f"Weather for {city}: {data}"
+        """
         # Check if user passed function directly instead of calling decorator.
         if callable(uri):
             raise TypeError(
@@ -342,7 +405,6 @@ class DurableMCP(fastmcp.FastMCP):
         return decorator
 
     def add_resource(self, resource: fastmcp.resources.Resource) -> None:
-        """Overrides `FastMCP.add_resource`."""
         # Where as `add_tool()` is a great insertion point,
         # `add_resource` gives us an already modified function which
         # does not pickle to the child process by default. This
@@ -359,6 +421,40 @@ class DurableMCP(fastmcp.FastMCP):
         title: str | None = None,
         description: str | None = None,
     ) -> Callable[[mcp.types.AnyFunction], mcp.types.AnyFunction]:
+        """Decorator to register a prompt.
+
+        Args:
+            name: Optional name for the prompt (defaults to function name)
+            title: Optional human-readable title for the prompt
+            description: Optional description of what the prompt does
+
+        Example:
+            @server.prompt()
+            def analyze_table(table_name: str) -> list[Message]:
+                schema = read_table_schema(table_name)
+                return [
+                    {
+                        "role": "user",
+                        "content": f"Analyze this schema:\n{schema}"
+                    }
+                ]
+
+            @server.prompt()
+            async def analyze_file(path: str) -> list[Message]:
+                content = await read_file(path)
+                return [
+                    {
+                        "role": "user",
+                        "content": {
+                            "type": "resource",
+                            "resource": {
+                                "uri": f"file://{path}",
+                                "text": content
+                            }
+                        }
+                    }
+                ]
+        """
         # Check if user passed function directly instead of calling decorator.
         if callable(name):
             raise TypeError(
@@ -380,7 +476,6 @@ class DurableMCP(fastmcp.FastMCP):
         return decorator
 
     def add_prompt(self, prompt: fastmcp.prompts.Prompt) -> None:
-        """Overrides `FastMCP.add_prompt`."""
         # Where as `add_tool()` is a great insertion point,
         # `add_prompt` gives us an already modified function which
         # does not pickle to the child process by default. This
@@ -389,6 +484,64 @@ class DurableMCP(fastmcp.FastMCP):
         # from `prompt`?
         raise NotImplementedError("Use `prompt()` decorator instead")
 
+    def tool(
+        self,
+        name: str | None = None,
+        title: str | None = None,
+        description: str | None = None,
+        annotations: mcp.types.ToolAnnotations | None = None,
+        structured_output: bool | None = None,
+    ) -> Callable[[mcp.types.AnyFunction], mcp.types.AnyFunction]:
+        """Decorator to register a tool.
+
+        Tools can optionally request a Context object by adding a parameter with the
+        Context type annotation. The context provides access to MCP capabilities like
+        logging, progress reporting, and resource access.
+
+        Args:
+            name: Optional name for the tool (defaults to function name)
+            title: Optional human-readable title for the tool
+            description: Optional description of what the tool does
+            annotations: Optional ToolAnnotations providing additional tool information
+            structured_output: Controls whether the tool's output is structured or unstructured
+                - If None, auto-detects based on the function's return type annotation
+                - If True, unconditionally creates a structured tool (return type annotation permitting)
+                - If False, unconditionally creates an unstructured tool
+
+        Example:
+            @server.tool()
+            def my_tool(x: int) -> str:
+                return str(x)
+
+            @server.tool()
+            def tool_with_context(x: int, ctx: Context) -> str:
+                ctx.info(f"Processing {x}")
+                return str(x)
+
+            @server.tool()
+            async def async_tool(x: int, context: Context) -> str:
+                await context.report_progress(50, 100)
+                return str(x)
+        """
+        # Check if user passed function directly instead of calling decorator
+        if callable(name):
+            raise TypeError(
+                "The @tool decorator was used incorrectly. Did you forget to call it? Use @tool() instead of @tool"
+            )
+
+        def decorator(fn: mcp.types.AnyFunction) -> mcp.types.AnyFunction:
+            self.add_tool(
+                fn,
+                name=name,
+                title=title,
+                description=description,
+                annotations=annotations,
+                structured_output=structured_output,
+            )
+            return fn
+
+        return decorator
+
     def add_tool(
         self,
         fn: mcp.types.AnyFunction,
@@ -398,7 +551,6 @@ class DurableMCP(fastmcp.FastMCP):
         annotations: mcp.types.ToolAnnotations | None = None,
         structured_output: bool | None = None,
     ) -> None:
-        """Overrides `FastMCP.add_tool`."""
         self._tools.append(
             Tool(
                 fn=fn,
@@ -410,10 +562,25 @@ class DurableMCP(fastmcp.FastMCP):
             )
         )
 
+    def application(self) -> Application:
+        """
+        Returns a Reboot `Application` for running the MCP tools,
+        resources, prompts, etc that were defined.
+        """
+        application = Application(servicers=self.servicers())
+
+        application.http.mount(
+            self._path,
+            factory=self.streamable_http_app_factory,
+        )
+
+        return application
+
     @property
     def streamable_http_app_factory(self):
         return functools.partial(
             _streamable_http_app,
+            self._log_level,
             self._path,
             self._resources,
             self._prompts,
@@ -422,13 +589,14 @@ class DurableMCP(fastmcp.FastMCP):
 
 
 def _streamable_http_app(
+    log_level: LogLevel,
     path: str,
     resources: list[Resource],
     prompts: list[Prompt],
     tools: list[Tool],
     external_context_from_request: Callable[[Request], ExternalContext],
 ):
-    mcp = fastmcp.FastMCP()
+    mcp = fastmcp.FastMCP(log_level=log_level)
 
     for resource in resources:
         mcp.resource(
@@ -493,10 +661,10 @@ def _wrap_tool(fn: mcp.types.AnyFunction) -> mcp.types.AnyFunction:
             issubclass(annotation, fastmcp.Context)
         ):
             raise TypeError(
-                "`DurableMCP` only injects `ToolContext` not `Context`")
+                "`DurableMCP` only injects `DurableContext` not `Context`")
         if (
             isinstance(annotation, type) and
-            issubclass(annotation, ToolContext)
+            issubclass(annotation, DurableContext)
         ):
             context_parameter_names.append(parameter_name)
         else:
@@ -504,22 +672,22 @@ def _wrap_tool(fn: mcp.types.AnyFunction) -> mcp.types.AnyFunction:
 
     wrapper_signature = signature.replace(parameters=wrapper_parameters)
 
-    async def wrapper(ctx: fastmcp.Context, *args, **kwargs):
-
-        context: WorkflowContext | None = _context.get()
-
-        assert context is not None
-
+    async def wrapper(
+        ctx: fastmcp.Context,
+        context: WorkflowContext,
+        *args,
+        **kwargs,
+    ):
         # To account for the lack of "intersection" types in
         # Python (which is actively being worked on), we instead
-        # create a new dynamic `ToolContext` instance that
+        # create a new dynamic `DurableContext` instance that
         # inherits from the instance of `WorkflowContext` that we
         # already have.
-        context.__class__ = ToolContext
+        context.__class__ = DurableContext
 
-        context = cast(ToolContext, context)
+        context = cast(DurableContext, context)
 
-        # Now we add the `ToolContextProtocol` properties.
+        # Now we add the `DurableContextProtocol` properties.
         context._event_aliases = set()
 
         context.session = DurableSession(ctx.session, context)
@@ -779,19 +947,49 @@ def _wrap_tool(fn: mcp.types.AnyFunction) -> mcp.types.AnyFunction:
 
             return fn(**dict(bound.arguments))
         except:
-            # TODO: print stack trace after we've fixed `memoize`
-            # effect validation bug.
-            #
-            # import traceback
-            # traceback.print_exc()
+            import traceback
+            traceback.print_exc()
             raise
 
-    setattr(wrapper, "__signature__", wrapper_signature)
-    wrapper.__name__ = fn.__name__
-    wrapper.__doc__ = fn.__doc__
+    async def wrapper_validating_effects(
+        ctx: fastmcp.Context,
+        *args,
+        **kwargs,
+    ):
+        context: WorkflowContext | None = _context.get()
 
-    return wrapper
+        assert context is not None
 
+        # Checkpoint the context since it is the `IdempotencyManager`.
+        checkpoint = context.checkpoint()
+
+        result = await wrapper(ctx, context, *args, **kwargs)
+
+        if context._effect_validation == EffectValidation.DISABLED:
+            return result
+
+        # Effect validation is enabled.
+        logger.info(
+            f"Re-running tool '{fn.__name__}' "
+            f"to validate effects. See {DOCS_BASE_URL}/develop/side_effects "
+            "for more information."
+        )
+
+        # Restore the context to the checkpoint we took above so we
+        # can re-execute `callable` as though it is being retried from
+        # scratch.
+        context.restore(checkpoint)
+
+        # TODO: check if `result` is different (we don't do this for
+        # other effect validation so we're also not doing it now).
+
+        return await wrapper(ctx, context, *args, **kwargs)
+
+    setattr(wrapper_validating_effects, "__signature__", wrapper_signature)
+    wrapper_validating_effects.__name__ = fn.__name__
+    wrapper_validating_effects.__doc__ = fn.__doc__
+
+    return wrapper_validating_effects
 
 
 class StreamableHTTPASGIApp:
@@ -822,7 +1020,7 @@ class StreamableHTTPASGIApp:
 
             # If this is a new session, i.e., we don't have an ID,
             # then generate one which we'll use as the session ID.
-            mcp_session_id = mcp_session_id or uuid4().hex
+            mcp_session_id = mcp_session_id or uuid7().hex
 
             headers = dict(request.headers)
             headers[STATE_REF_HEADER] = f"rbt.mcp.v1.Session:{mcp_session_id}"
@@ -842,6 +1040,10 @@ class StreamableHTTPASGIApp:
                         str(request.url),
                         headers=headers,
                         content=await request.body(),
+                        # Don't worry about timing out, this might be
+                        # a long-lived `GET` for server sent events
+                        # streaming.
+                        timeout=None,
                     ) as upstream:
                         # Create a generator to yield chunks from the
                         # upstream response.
@@ -870,14 +1072,6 @@ class StreamableHTTPASGIApp:
 
         mcp_session_id = request.headers.get(MCP_SESSION_ID_HEADER)
 
-        if (
-            mcp_session_id is not None and
-            mcp_session_id in self._http_transports
-        ):
-            transport = self._http_transports[mcp_session_id]
-            await transport.handle_request(scope, receive, send)
-            return
-
         # If this is a new session, i.e., `mcp_session_id is None`, we
         # need to use the session ID we already generated so all
         # requests for it will get routed to this consensus.
@@ -887,15 +1081,110 @@ class StreamableHTTPASGIApp:
             session_id = StateRef.from_maybe_readable(session_ref).id
             mcp_session_id = session_id
 
+        session = Session.ref(mcp_session_id)
+
+        _is_vscode: bool | None = None
+
+        async def is_vscode():
+            """Returns true if this session client is Visual Studio Code."""
+            nonlocal _is_vscode
+            if _is_vscode is None:
+                backoff = Backoff(max_backoff_seconds=2)
+                while _is_vscode is None:
+                    try:
+                        # TODO: not using `session.reactively().get()`
+                        # because it doesn't properly propagate
+                        # `Session.GetAborted`.
+                        response = await session.get(context)
+
+                        # Need to wait until session has been initialized,
+                        # which is once `client_info` is populated.
+                        if not response.HasField("client_info"):
+                            await backoff()
+                            continue
+
+                        # Technically `name` is required but at least the
+                        # MCP SDK doesn't validate it via Pydantic, but
+                        # Visual Studio Code always seems to include its
+                        # name, so if we don't have a name it is not
+                        # Visual Studio Code.
+                        if response.client_info.HasField("name"):
+                            _is_vscode = (
+                                response.client_info.name == "Visual Studio Code"
+                            )
+                        else:
+                            _is_vscode = False
+                    except Session.GetAborted as aborted:
+                        if type(aborted.error) == StateNotConstructed:
+                            await backoff()
+                            continue
+                        raise
+            assert _is_vscode is not None
+            return _is_vscode
+
+        # If this is a GET and the client is Visual Studio Code always
+        # ensure it has a 'last-event-id' so that it always replays
+        # from the aggregate stream.
+        if request.method == "GET":
+            if "last-event-id" not in request.headers:
+                if await is_vscode():
+                    # Modify headers to always include a
+                    # 'last-event-id' so that we'll always
+                    # replay from the aggregate stream.
+                    mutable_headers = MutableHeaders(scope=scope)
+                    mutable_headers[
+                        "last-event-id"
+                    ] = "VSCODE_INITIAL_GET_LAST_EVENT_ID"
+                    scope["headers"] = mutable_headers.raw
+                    request = Request(scope, receive)
+
+        drop = False
+        async def post_send(message) -> None:
+            """
+            Helper that drops Visual Studio Code events from POST
+            requests since we send all events over GET.
+
+            We do this here so that the events are still sent
+            through the MCP SDK so everything gets cleaned up
+            correctly.
+            """
+            nonlocal drop
+            if message["type"] == "http.response.start":
+                if message["status"] == 200:
+                    for key, value in message["headers"]:
+                        if (
+                            key == b"content-type" and
+                            value == b"text/event-stream" and
+                            await is_vscode()
+                        ):
+                            # We want to drop Visual Studio Code
+                            # streams because we send everything
+                            # through the GET.
+                            drop = True
+                            break
+                # But always send "http.response.start" so Visual
+                # Studio Code at least gets the initial 200.
+                await send(message)
+            elif not drop:
+                # Send on messages for things like 202 Accepted.
+                await send(message)
+
+        if mcp_session_id in self._http_transports:
+            transport = self._http_transports[mcp_session_id]
+            await transport.handle_request(
+                scope,
+                receive,
+                send if request.method == "GET" else post_send,
+            )
+            return
+
         http_transport = StreamableHTTPServerTransport(
             mcp_session_id=mcp_session_id,
             is_json_response_enabled=False,
-            event_store=DurableEventStore(context),
+            event_store=DurableEventStore(context, mcp_session_id),
             security_settings=None,
         )
 
-        session = Session.ref(mcp_session_id)
-
         self._http_transports[mcp_session_id] = http_transport
 
         started = asyncio.Event()
@@ -909,6 +1198,7 @@ class StreamableHTTPASGIApp:
                 writer_tasks: list[asyncio.Task] = []
 
                 async def reader():
+
                     async for message in read_stream:
                         # TODO: we can't pickle `request_context`
                         # which is a `starlette.requests.Request`,
@@ -937,11 +1227,19 @@ class StreamableHTTPASGIApp:
                             async def writer():
                                 async for message, _ in replay(
                                     context,
-                                    stream_id=str(request_id),
+                                    session_id=session.state_id,
+                                    request_id=request_id,
                                 ):
                                     await write_stream.send(message)
 
-                            writer_tasks.append(asyncio.create_task(writer()))
+                            writer_task = asyncio.create_task(writer())
+
+                            writer_tasks.append(writer_task)
+
+                            def done(task):
+                                writer_tasks.remove(task)
+
+                            writer_task.add_done_callback(done)
 
                 reader_task = asyncio.create_task(reader())
 
@@ -979,4 +1277,8 @@ class StreamableHTTPASGIApp:
 
         await started.wait()
 
-        await http_transport.handle_request(scope, receive, send)
+        await http_transport.handle_request(
+            scope,
+            receive,
+            send if request.method == "GET" else post_send,
+        )
